#include "pipe.h"
#include "node.h"
#include <fstream>
#include <cstdlib>
#include <cmath>
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <armadillo>
#include <time.h>
#include <ctime>
#include <stdio.h>      /* printf, NULL */
#include <stdlib.h>
#define maxBubbles 15
#define ERR 1.0e-14
#define clamp(x,a,b) (x < a ? a : x > b ? b : x)

using namespace std;
using namespace arma;

class pipe;
class node;

int M = 30;
int N = 30;
int yLen = M;
int nNodes = N*M;
int nPipes = N*M*3/2;
double Theta = 0;

double PI = 4.0*atan(1.0);
double sTens = 3.0;         // surface tension: units = dyn/mm == 10 mN/m
double muNON = 0.01;        // viscosity: units = Po == 10 Pa*s
double muWET = 0.01;     
double radMin = 0.1;
double radMax = 0.4;
double totalT = 0;
double pressure; //global pressure gradient

double dt;
double poreVolume;
double saturation;

int iterations = 50000; 
int numPoints = 10; //number of saturation points between 0 and 1


mat A = zeros(nNodes,nNodes); //Cunductance matrix when solvng pressure
mat Ainv = zeros(nNodes,nNodes);
vec B = zeros(nNodes); //Containing capillary pressurers and boundary condtitions 

node* nodes = new  node[nNodes]; 
Pipe* pipes = new Pipe[nPipes];

//armadillo vectors and matrixes containig simulation data
mat bubbleStartOut = zeros(nPipes*400,5); //start position of bubbles for max 400 timesteps
mat bubbleStopOut = zeros(nPipes*400,5); // end position of all bubbles max 400 timesteps
vec TotFlow = zeros(iterations); //Total flow for one realization
vec FracFlow = zeros(iterations); // Fractional flow for one realization
mat avgFlow =zeros(numPoints,11); //Average flow value at steady state for mutilple realizations
mat avgFrac = zeros(numPoints,11); // Average fractional flow value at steady state for multiple realizations
vec Ca = zeros(iterations); // Capillary numbers for one realization
mat avgCa = zeros(numPoints,11); // Average capillary number at steady state for multiple realizations
vec timevec = zeros(iterations); //Timevalues for one realization


void initializeNodesAndPipes();
void getFlow();
void getPoreVolume();
double calcDeltaT();
void updateBubbles();
void setOutput(int i, int K);
void solvePressure();
void meashureFlow(int i);
double totSaturation();


void initializeNodesAndPipes(){
	for (int i = 0; i<nPipes; i++){
		pipes[i].index = i;
		pipes[i].flow = 1; //ta bort?
		pipes[i].nBubbles = 0;
		pipes[i].theta = Theta;
		pipes[i].swapped = false;
	}
	for (int i = 0; i<nNodes; i++){
		nodes[i].index = i;
		nodes[i].setConnections(pipes,nodes,N,M);
		nodes[i].pressure = 0;
	}
	for(int i = 0; i<(int)(nPipes*saturation); i++){
		pipes[i].bubbleStop[0] = 1.0;
		pipes[i].bubbleStart[0] = 0.0;
		pipes[i].nBubbles = 1;
	}
	for(int i = 0; i<nPipes; i++){
		pipes[i].getCapBoundaries();
		pipes[i].calcPcandMobility();
		pipes[i].getFlow(pressure);
	}
}

void getPoreVolume(){
	poreVolume = 0;
	for(int i = 0; i<nPipes; i++){
		poreVolume += pipes[i].area;
	}
}
void getFlow(){
	for (int i = 0; i<nPipes; i++){
		pipes[i].getFlow(pressure);
	}
}


double calcDeltaT() {
//calculate adapitve timestep
    double velMax = 0.0;
    int velMaxPos = -1;
    for (int i = 0; i < nPipes; i++) {
        double vel = pipes[i].flow / pipes[i].area; // volume-normalized velocity
        if (saturation == 0.0 || saturation == 1.0) {
            if (velMax < fabs(vel)) {
                velMax = fabs(vel);
                velMaxPos = i;
            }
        } else {
            if (velMax < fabs(vel) && pipes[i].nBubbles > 0) {
                velMax = fabs(vel);
                velMaxPos = i;
            }
        }
    }
    if (velMaxPos == -1) {
        printf("No maximum velocity found.\n"); fflush(stdout);
        exit(1);
    }
    return 0.1 / velMax * 1.0;
}

double totSaturation(){
	//returns the total saturation of the system
	double totSat = 0;
	for (int i = 0; i < nPipes; i++){
		totSat += pipes[i].calcLinkSaturation()*pipes[i].area;
	}
	return totSat;
}

void updateBubbles(){
		for (int i = 0; i<nNodes; i++){
			nodes[i].nodeFrac = 0;
		}
		dt = calcDeltaT();
		for (int i = 0; i<nPipes; i++){
			pipes[i].moveBubble(dt);
		}
		for (int i = 0; i<nNodes; i++){
			nodes[i].updateNodeFrac();
		}
		for (int i = 0; i<nPipes; i++){
			pipes[i].distributeBubbles();
			
		}
		for (int i = 0; i<nNodes; i++){
			nodes[i].bubbleAdjust(pipes);
		}
		for (int i = 0; i<nPipes; i++){
			pipes[i].killBubbles();
		}
}

void setOutput(int i, int K,int k){
	//Write bubble positions to an armadillo vector used for visualization
	bool swappedd;
	if(i%K==0){
		k++;
		for (int i = 0; i<nPipes;i++){
			swappedd = false;
			if((!pipes[i].swapped && pipes[i].nodeN->verticalPipe->index != pipes[i].index) || (pipes[i].swapped && pipes[i].nodeN->verticalPipe->index == pipes[i].index)){
				swappedd = true;
				pipes[i].Swap();
			}
			for(int j = 0; j<5; j++){
				bubbleStartOut(i+N*M/2*3*(k-1),j) = pipes[i].bubbleStart[j];
				bubbleStopOut(i+N*M/2*3*(k-1),j) = pipes[i].bubbleStop[j];
			}
			if(swappedd){
				pipes[i].Swap();
			}
		}
	}
}

void setAinv(){
	A(0,0) = 1;
	for(int i = 1; i<nNodes; i++){
		A(i,i) 				= - nodes[i].rightPipe->mobility - nodes[i].leftPipe->mobility - nodes[i].verticalPipe->mobility;
		A(i,nodes[i].verticalNode->index) 	= nodes[i].verticalPipe->mobility;
		A(i,nodes[i].leftNode->index) 		=  nodes[i].leftPipe->mobility;
		A(i,nodes[i].rightNode->index) 		=  nodes[i].rightPipe->mobility;
	}
	
	Ainv = inv(A);
}

void solveConstA(){
	//find the pressure when A is constant
	int n,m;
	double b[nNodes];
	B(0) = 0;
	for(int i = 1; i<nNodes; i++){
		B(i) = 0;
		for(int j = 0; j < 3; j++){
			if(j == 0){
				m = nodes[i].leftPipe->index;
			}else if (j == 1){
				m = nodes[i].rightPipe->index;
			}
			else{
				m = nodes[i].verticalPipe->index;
			}
			
			B(i) -= pipes[m].mobility*(-pipes[m].Pc+pressure*pipes[m].boundary)*(nodes[i].upper? 1: -1);
		}
	}
	vec P = Ainv*B;
	for(int i = 0; i<nNodes; i++){
		nodes[i].pressure = P(i);
	}
}

void solvePressure(){
	//conjugate gradient method for general case of different viscocities
	double p[nNodes];
	double r[nNodes];
	double ap[nNodes];
	double rpn = 0.0;
	node c_node;
	Pipe link;
	for (int i = 0; i < nNodes; i++){
		p[i] = 0;
		for(int j = 0; j<3; j++){
		if(j == 0){
        		link = *nodes[i].leftPipe;
        		c_node = *nodes[i].leftNode;}
        	else if (j == 1){
        		link = *nodes[i].rightPipe;
        		c_node = *nodes[i].rightNode;}
        	else{
        		link = *nodes[i].verticalPipe;
        		c_node = *nodes[i].verticalNode;}
			
			p[i] -= link.mobility*((c_node.pressure - nodes[i].pressure)+ (-link.Pc + pressure*link.boundary)*(nodes[i].upper? 1: -1) );
			}
		
		r[i] = p[i];
		rpn += r[i]*r[i];
	}
	
	int v = 0;
	while (rpn >= ERR){
		v++;
		double rps = rpn;
		double am = 0.0;
		for(int i = 0; i < nNodes; i++){
			double api = nodes[i].leftPipe->mobility * (p[nodes[i].leftNode->index] - p[i]) +
			             nodes[i].rightPipe->mobility * (p[nodes[i].rightNode->index] - p[i]) +
			             nodes[i].verticalPipe->mobility * (p[nodes[i].verticalNode->index] - p[i]);
			ap[i] = api;
			am   += p[i]*api;
		}
		am = rps/am;
		rpn = 0.0;
		for(int i = 0; i<nNodes; i++){
			nodes[i].pressure += am * p[i];
			r[i] -= am * ap[i];
			rpn += r[i] * r[i];
		}
		
		double bm = rpn / rps;
		for (int i = 0; i < nNodes; i++){
			p[i] = r[i] + bm*p[i];
		}
	}
}

void meashureFlow(int i){
	double totalFlow = 0.0;
	double start;
	double stop;
	double totalFlowNW = 0.0;
	double volNW = 0.0;
	double volTL = 0.0;
	double C_a = 0.0;
	double cArea = 0.0;
	for (int j = N; j<3*N/2; j++){
		totalFlow += pipes[j].lenTL*pipes[j].area*(pipes[j].swapped?-1:1);
		cArea += pipes[j].area;
	}
	Ca(i) = totalFlow/cArea*muNON/sTens/dt;
	TotFlow(i) = totalFlow/dt;
	
	for (int j = 0; j<nPipes; j++){
		if(pipes[j].nodeN->verticalPipe->index == pipes[j].index){
			start = 0.5;
			stop = 0.5+pipes[j].lenTL;
			for (int l = 0; l < pipes[j].nBubbles; l++) {
           			volNW += (clamp(pipes[j].bubbleStop[l], start, stop) - clamp(pipes[j].bubbleStart[l], start, stop))*pipes[j].area*(pipes[j].swapped?-1:1)/dt;
		 	}
		 	volTL += pipes[j].lenTL*pipes[j].area/dt*(pipes[j].swapped?-1:1);
		}
	
	}
	FracFlow(i) = volNW/volTL;
}




int main(){
	srand (time(NULL)); //set radom seed
	double r_mean = 0.25;
	pressure =0.7* yLen * 2 * sTens / r_mean; //global pressure gradient
	int iter2 = 0;
	for(Theta = 0.0; Theta <=PI/2.0; Theta += PI/2.0/10.0){ //looping over wetting angles
		int iter1 = 0;
		for(saturation = 0.0; saturation <= 1.00; saturation += 1/((double)numPoints-1)){ //looping over saturation
		cout<<"iteration1: " << iter1<<endl;
		cout<<"saturation: "<<saturation<<endl;
		cout<<endl;
		initializeNodesAndPipes();	
		getPoreVolume();
	
	
		//timers
		clock_t pressure;
		clock_t update;
		clock_t flow;
		double dPressure = 0;
		double dUpdate = 0;
		double dFlow = 0;
	
		setAinv();
		dPressure = 0;

		int i = 0;
		int k = 0;
		while (i < iterations){
			//main loop
			for(int j = 0; j<nPipes; j++){
				pipes[j].calcPcandMobility();
			}
			pressure = clock();
			solveConstA();
			dPressure += (clock()-pressure)/(double) CLOCKS_PER_SEC;
			flow = clock();
			getFlow();
			dFlow += (clock()-flow)/(double) CLOCKS_PER_SEC;
			update = clock();
			updateBubbles();
			dUpdate += (clock()-update)/(double) CLOCKS_PER_SEC;
			meashureFlow(i);
			setOutput(i,25,k); //write bubble positions to vector	
			totalT += dt;
			timevec(i) = totalT;
			i++;
		}
		cout<<"time "<<totalT<<endl;
	
		//output the flow and coordinates of each tube.
		mat tubeData = zeros(5,N*M/2*3); 
		for (int i = 0; i<N*M/2*3; i++){
			tubeData(0,i) = pipes[i].flow;
			tubeData(1,i) = pipes[i].x1;
			tubeData(2,i) = pipes[i].y1;
			tubeData(3,i) = pipes[i].x2;
			
			tubeData(4,i) = pipes[i].y2;
		}
		tubeData.save("Output.txt",raw_ascii);
		for(int i = iterations/2; i<iterations; i++){
			avgFlow(iter1,iter2) += TotFlow(i);
			avgFrac(iter1,iter2) += FracFlow(i);
			avgCa(iter1,iter2) += Ca(i);
		}
		avgFlow(iter1,iter2) = avgFlow(iter1,iter2)/((double)iterations/2);
		avgFrac(iter1,iter2) = avgFrac(iter1,iter2)/((double)iterations/2);
		avgCa(iter1,iter2) = avgCa(iter1,iter2)/((double)iterations/2);
	
		//print simulation data
		cout<<"avgFlow: "<< endl <<avgFlow<<" avgFrac: "<< endl <<avgFrac<<endl<<"avgCa "<<avgCa<<endl;
		cout<<"tPressure "<<dPressure<<" tFlow "<<dFlow<<" tUpdate "<<dUpdate<<endl;
		iter1++;
		}
	iter2++;
	}
	//Save datasets
	//bubbleStartOut.save("bStart2.txt",raw_ascii);
	//bubbleStopOut.save("bStop2.txt",raw_ascii);
	avgFlow.save("Q_Circ_Circ.txt",raw_ascii);
	timevec.save("timedt02Long.txt",raw_ascii);
	TotFlow.save("Flowdt02Long.txt",raw_ascii);
	avgFrac.save("F_Circ_Circ.txt",raw_ascii);
	avgCa.save("Ca_Circ_Circ.txt",raw_ascii);
return 0;
}
