#include "node.h"
#include <fstream>
#include <cstdlib>
#include <cmath>
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <armadillo>
#include <time.h>
#define SIGN(x) (x < 0) ? -1 : (x > 0)
#define TOL 1.0e-12

using namespace std;
using namespace arma;

void node::setConnections(Pipe pipes[],node nodes[],int N,int M){
	//algorithm to fin hexagonal connections based on index
	n = (index)%(N);
	m = floor((index)/(N));

	x = (n+1)*sqrt(3)/2;
	y = ((m%2)?(n%2)*0.5:(1-(n%2))*0.5)+1.5*m;
	
	leftPipe  = &pipes[(m*N*3/2+n)-1+(n==0)*(N)];
	leftNode  = &nodes[index-1+(n==0)*(N)];
	
	rightPipe = &pipes[m*N*3/2+n];
	rightNode = &nodes[index+1-(n==(N-1))*(N)];
	
	double above = m*N*3/2+N+(double)(n+1+((m+1)%(2)))/2-1;
	double below = m*N*3/2-(double)(N-n+((m+1)%(2)))/2+(m==0)*M*N*3/2;
	if (ceil(above) == above){
		verticalPipe = &pipes[(int)above];
		verticalNode = &nodes[index+N];
		upper = true;
		if (m == (M-1)){
			verticalNode = &nodes[index+N-(m==(M-1))*N*M];
			verticalPipe->boundary = 1; 
		}
	}
	else{
		verticalPipe = &pipes[(int)below];
		verticalNode = &nodes[index-N];
		upper = false;
		if(m==0){
			verticalNode = &nodes[index-N+N*M];
			verticalPipe->boundary = 1;
		}
	}
	
	if(upper){
		leftPipe->nodeN = this;
		rightPipe->nodeN = this;
		verticalPipe->nodeN = this;
	}else{
		leftPipe->nodeP = this;
		rightPipe->nodeP = this;
		verticalPipe->nodeP = this;
	}
	
	leftPipe->setCoordinates(x,y);
	rightPipe->setCoordinates(x,y);
	verticalPipe->setCoordinates(x,y);
}

void node::updateNodeFrac(){
	nodeFrac = nodeFluxNW/(nodeFluxTL+10e-50);
	nodeFluxNW = 0.0;
	nodeFluxTL = 0.0;
}

void node::bubbleAdjust(Pipe* pipes) {
	double dx1s[3] = { 1.0, 1.0, 1.0 }; // sizes of first W  bubble in each neighouring tube
        double dx2s[3] = { 0.0, 0.0, 0.0 }; // sizes of first NW bubble
        int typeFirst[3] = { 0,0,0 }; // 1 == NW, 0 == W
        int links[3];
        int link;
        int linksId[3] = { 0,0,0 };
        int ctr = 0;
        bool isSmallW = false;
        bool isSmallNW = false;
        int ctr2 = 0;
        for(int i = 0; i < 3; i++){
        	if(i == 0){
        		link = leftPipe->index;
        	}else if (i == 1){
        		link = rightPipe->index;
        	}
        	else{
        		link = verticalPipe->index;
        	}

        	int bubble = (pipes[link].nodeN->index == index ? 0 : pipes[link].nBubbles-1);
        	dx1s[i] = (pipes[link].nodeN->index == index ? pipes[link].bubbleStart[bubble] : 1.0 - pipes[link].bubbleStop[bubble]); //switched
        	dx2s[i] = pipes[link].bubbleStop[bubble] - pipes[link].bubbleStart[bubble];
        	
        	if(dx1s[i] < 10e-12){
        		
        		if(dx2s[i] < pipes[link].radius && dx2s[i]> 10e-12){
        			typeFirst[ctr] = 1;
        			links[ctr] = link;
        			linksId[ctr++] = i;
        			isSmallNW = true;
        		}
        		}
        	else{
        		if (dx1s[i] <pipes[link].radius && dx1s[i] > 10e-12) {
        			links[ctr] = link;
        			linksId[ctr++] = i;
        			isSmallW = true;
        		}
        	}
        }
        
            if (isSmallW && isSmallNW) {
            double sumW = 0.0;
            double sumNW = 0.0;
            for (int i = 0; i < ctr; i++) {
                if (typeFirst[i]) {
                    sumNW += dx2s[linksId[i]]*pipes[links[i]].area;
                } else {
                    sumW += dx1s[linksId[i]]*pipes[links[i]].area;
                }
            }
            double rem = 0.0;
            double add = 0.0;
            double minVol = min(sumNW, sumW);

            for (int i = 0; i < ctr; i++) {
                int link = links[i];
                if (typeFirst[i]) {
                    // remove NW from tubes with small NW
                    if (pipes[link].nodeP->index == index) {
                        pipes[link].bubbleStart[pipes[link].nBubbles-1] += dx2s[linksId[i]]*minVol/sumNW;
                        rem +=  (dx2s[linksId[i]]*minVol/sumNW)*pipes[link].area;
                    } else {
                        pipes[link].bubbleStop[0] -= dx2s[linksId[i]]*minVol/sumNW;
                        rem +=  (dx2s[linksId[i]]*minVol/sumNW)*pipes[link].area;
                    }                     
                } else {
                    // add NW to tubes with small W
                    if (pipes[link].nodeP->index == index) {
                        pipes[link].bubbleStop[pipes[link].nBubbles-1] += dx1s[linksId[i]] * minVol/sumW;
                        add +=  (dx1s[linksId[i]] * minVol/sumW)*pipes[link].area;
                    } else {
                        pipes[link].bubbleStart[0] -= dx1s[linksId[i]] * minVol/sumW;
                        add +=  (dx1s[linksId[i]] * minVol/sumW)*pipes[link].area;
                    }     
                }
            }
        } else if (isSmallNW && ctr == 1) {
            int bubble = 0;
            if (nodeFrac < TOL && pipes[links[0]].nodeN->index == index) {
                pipes[links[0]].bubbleStart[bubble] += pipes[links[0]].lenTL;
                pipes[links[0]].bubbleStop[bubble] += pipes[links[0]].lenTL;
            }
        }
        
        
          else if ((ctr == 1 && isSmallW)) {
        	int bubble = 0;// = (pipes[links[0]].nodeP->index == index ? pipes[links[0]].nBubbles - 1 : 0);
        	if (nodeFrac > 1 - TOL &&  pipes[links[0]].nodeN->index == index) {
                	pipes[links[0]].bubbleStart[bubble] += pipes[links[0]].lenTL;
                	for(int i = 15 -1 ; i > 0; i--){
     					pipes[links[0]].bubbleStart[i] = pipes[links[0]].bubbleStart[i-1];
     					pipes[links[0]].bubbleStop[i] = pipes[links[0]].bubbleStop[i-1];	
     				}
     				pipes[links[0]].nBubbles++;
     				pipes[links[0]].bubbleStart[0] = 0.0;
     				pipes[links[0]].bubbleStop[0]  = pipes[links[0]].lenTL;
            	}
        }
}
